/*
 * nfsmount.c -- Linux NFS mount
 * Copyright (C) 1993 Rick Sladkey <jrs@world.std.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Wed Feb  8 12:51:48 1995, biro@yggdrasil.com (Ross Biro): allow all port
 * numbers to be specified on the command line.
 *
 * Fri, 8 Mar 1996 18:01:39, Swen Thuemmler <swen@uni-paderborn.de>:
 * Omit the call to connect() for Linux version 1.3.11 or later.
 *
 * Modified by Olaf Kirch for new NFS code, plus NFSv3 stuff.
 * Note: this code will not compile on older kernels anymore.
 */

#define NFS_MOUNT_DEBUG
#include <stdio.h>
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <rpc/pmap_clnt.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>

#include "sundries.h"
#include "nfsmount.h"

#define _LINUX_SOCKET_H	/* Avoid clash with glibc 2 */
#include <linux/fs.h>
#include <linux/nfs_fs.h>
#include <linux/nfs_mount.h>

/* Hack to find out whether the kernel's been patched for NFSv3 support */
#ifdef NFSERR_JUKEBOX
# define HAVE_NFSV3
#endif

#if !defined(__GLIBC__) || __GLIBC__ < 2
static char *
strndup (char *str, int n)
{
	char	*ret;

	ret = malloc(n + 1);
	if (ret == NULL) {
		perror("malloc");
		return (NULL);
	}
	strncpy(ret, str, n);
	ret[n] = '\0';
	return (ret);
}
#endif

static char *nfs_strerror(int stat);

#if 0
static int
linux_version_code(void) {
	struct utsname my_utsname;
	int release = 0;
  
	if (0 == uname(&my_utsname)) {
	  release = 65536 * atoi(strtok(my_utsname.release, "."))
	           +  256 * atoi(strtok(NULL, "."))
		   +        atoi(strtok(NULL, "."));
	}
	return release;
}
#endif

/*
 * Parse all those nifty mount options.
 */
static struct nfs_mount_data	data;
static const char *		mnthost;
static int			nfsport, nfsprog, nfsvers, nfsproto;
static int			mntport, mntprog, mntvers;
static int			retry;

static int
parse_opts(char *options)
{
	char	*opt, *opteq;
	int	val, bg = 0, soft = 0, intr = 0, posix = 0, nocto = 0,
		noac = 0, tcp = 0, nolock = 0, secure = 0;

	/*
	 * Set default options.
	 * rsize/wsize/bsize are left 0 in order to let the kernel decide.
	 * Timeo is filled in after we know whether it'll be TCP or UDP.
	 */
	memset(&data, 0, sizeof(data));
	data.version  = NFS_MOUNT_VERSION;
	data.retrans  = 3;
	data.acregmin = 3;
	data.acregmax = 60;
	data.acdirmin = 30;
	data.acdirmax = 60;
	data.fd       = -1;		/* obsolete */

	mntprog  = MOUNTPROG;
	mntvers  = 0;
	mntport  = 0;
	nfsprog  = NFS_PROGRAM;
	nfsvers  = NFS_VERSION;
	nfsproto = IPPROTO_UDP;
	nfsport  = 0;
	retry    = 10000;

	for (opt = strtok(options, ","); opt; opt = strtok(NULL, ",")) {
		if ((opteq = strchr(opt, '='))) {
			val = atoi(opteq + 1);	
			*opteq = '\0';
			if (!strcmp(opt, "rsize"))
				data.rsize = val;
			else if (!strcmp(opt, "wsize"))
				data.wsize = val;
			else if (!strcmp(opt, "timeo"))
				data.timeo = val;
			else if (!strcmp(opt, "retrans"))
				data.retrans = val;
			else if (!strcmp(opt, "acregmin"))
				data.acregmin = val;
			else if (!strcmp(opt, "acregmax"))
				data.acregmax = val;
			else if (!strcmp(opt, "acdirmin"))
				data.acdirmin = val;
			else if (!strcmp(opt, "acdirmax"))
				data.acdirmax = val;
			else if (!strcmp(opt, "actimeo")) {
				data.acregmin = val;
				data.acregmax = val;
				data.acdirmin = val;
				data.acdirmax = val;
			}
			else if (!strcmp(opt, "retry"))
				retry = val;
			else if (!strcmp(opt, "port"))
				nfsport = val;
			else if (!strcmp(opt, "mountport"))
			        mntport = val;
			else if (!strcmp(opt, "mounthost"))
			        mnthost=strndup(opteq+1,
						  strcspn(opteq+1," \t\n\r,"));
			else if (!strcmp(opt, "mountprog"))
				mntprog = val;
			else if (!strcmp(opt, "mountvers"))
				mntvers = val;
			else if (!strcmp(opt, "nfsprog"))
				nfsprog = val;
			else if (!strcmp(opt, "nfsvers"))
				nfsvers = val;
			else if (!strcmp(opt, "namlen"))
				data.namlen = val;
			else if (!strcmp(opt, "addr"))
				/* ignore */;
			else {
				printf("unknown nfs mount parameter: "
				       "%s=%d\n", opt, val);
				return 0;
			}
		}
		else {
			val = 1;
			if (!strncmp(opt, "no", 2)) {
				val = 0;
				opt += 2;
			}
			if (!strcmp(opt, "bg")) 
				bg = val;
			else if (!strcmp(opt, "fg")) 
				bg = !val;
			else if (!strcmp(opt, "soft"))
				soft = val;
			else if (!strcmp(opt, "hard"))
				soft = !val;
			else if (!strcmp(opt, "intr"))
				intr = val;
			else if (!strcmp(opt, "posix"))
				posix = val;
			else if (!strcmp(opt, "cto"))
				nocto = !val;
			else if (!strcmp(opt, "ac"))
				noac = !val;
			else if (!strcmp(opt, "tcp"))
				tcp = val;
			else if (!strcmp(opt, "udp"))
				tcp = !val;
			else if (!strcmp(opt, "lock"))
				nolock = !val;
			else if (!strcmp(opt, "secure"))
				secure = val;
			else {
				printf("unknown nfs mount option: "
				       "%s%s\n", val ? "" : "no", opt);
				return 0;
			}
		}
	}
	data.flags = (soft ? NFS_MOUNT_SOFT : 0)
		| (intr ? NFS_MOUNT_INTR : 0)
		| (posix ? NFS_MOUNT_POSIX : 0)
		| (nocto ? NFS_MOUNT_NOCTO : 0)
		| (noac ? NFS_MOUNT_NOAC : 0)
		| (tcp ? NFS_MOUNT_TCP : 0)
		| (nolock? NFS_MOUNT_NONLM : 0)
		| (secure? NFS_MOUNT_SECURE : 0);

	if (nfsvers == 3) {
#ifndef HAVE_NFSV3
		fprintf(stderr, "NFSv%d not supported!\n", nfsvers);
		return 0;
#else
		data.flags |= NFS_MOUNT_VER3;
		if (mntvers < 3)
			mntvers = 3;
#endif
	} else if (nfsvers == 2) {
		if (mntvers > 2)
			mntvers = 1;
	} else {
		fprintf(stderr, "NFSv%d not supported!\n", nfsvers);
		return 0;
	}

	if (data.flags & NFS_MOUNT_TCP)
		nfsproto = IPPROTO_TCP;

	/* Adjust options if none specified */
	if (!data.timeo)
		data.timeo = tcp? 70 : 7;

#ifdef NFS_MOUNT_DEBUG
	printf("rsize = %d, wsize = %d, timeo = %d, retrans = %d\n",
		data.rsize, data.wsize, data.timeo, data.retrans);
	printf("acreg (min, max) = (%d, %d), acdir (min, max) = (%d, %d)\n",
		data.acregmin, data.acregmax, data.acdirmin, data.acdirmax);
	printf("port = %d, tcp = %d, bg = %d, retry = %d, flags = %.8x\n",
		nfsport, (data.flags & NFS_MOUNT_TCP) != 0, bg, retry,
		data.flags);
	printf("mountprog = %d, mountvers = %d, nfsprog = %d, nfsvers = %d\n",
		mntprog, mntvers, nfsprog, nfsvers);
	printf("soft = %d, intr = %d, posix = %d, nocto = %d, noac = %d\n",
		(data.flags & NFS_MOUNT_SOFT) != 0,
		(data.flags & NFS_MOUNT_INTR) != 0,
		(data.flags & NFS_MOUNT_POSIX) != 0,
		(data.flags & NFS_MOUNT_NOCTO) != 0,
		(data.flags & NFS_MOUNT_NOAC) != 0);
#endif

	return 1;
}

/*
 * Obtain the address for a given hostname
 */
static int
get_servaddr(const char *hostname, struct sockaddr_in *sin)
{
	struct hostent	*hp;

	sin->sin_family = AF_INET;
	if (hostname[0] >= '0' && hostname[0] <= '9') {
		sin->sin_addr.s_addr = inet_addr(hostname);
	} else if ((hp = gethostbyname(hostname)) == NULL) {
		fprintf(stderr, "mount: can't get address for %s\n", hostname);
		return 0;
	} else {
		memcpy(&sin->sin_addr, hp->h_addr, hp->h_length);
	}

	return 1;
}

static CLIENT *
get_mount_client(struct sockaddr_in *sin, int port, int prog, int vers)
{
	struct timeval	timeout;
	int		msock;
	CLIENT		*clnt;

#if 0
	/* First try to create a TCP client */
	sin->sin_port = htons(port);
	msock = RPC_ANYSOCK;
	if ((clnt = clnttcp_create(sin, prog, vers, &msock, 0, 0)) != NULL) {
#ifdef NFS_MOUNT_DEBUG
		printf("using MNTv%d/TCP for mount deamon\n", vers);
#endif
		return clnt;
	}
#endif

	/* Try again using UDP */
	sin->sin_port = htons(port);
	msock = RPC_ANYSOCK;
	timeout.tv_sec = 3;
	timeout.tv_usec = 0;
	if (!(clnt = clntudp_create(sin, prog, vers, timeout, &msock)))
		return NULL;

#ifdef NFS_MOUNT_DEBUG
	printf("using MNTv%d/UDP for mount deamon\n", vers);
#endif
	return clnt;
}

/*
 * Obtain NFSv2 file handle from server
 */
static int
get_rootfh_nfs2(CLIENT *clnt, const char *hostname,
			      const char *dirname, struct nfs_fh *fh)
{
	struct timeval	timeout;
	enum clnt_stat	clnt_stat;
	struct fhstatus	status;

	timeout.tv_sec  = 20;
	timeout.tv_usec = 0;

	/* try to mount hostname:dirname */
	clnt_stat = clnt_call(clnt, MOUNTPROC_MNT,
		(xdrproc_t) xdr_dirpath,  &dirname,
		(xdrproc_t) xdr_fhstatus, &status,
		timeout);

	if (clnt_stat != RPC_SUCCESS) {
		clnt_perror(clnt, "rpc mount");
		return 0;
	}
	if (status.fhs_status != 0) {
		fprintf(stderr,
			"mount: %s:%s failed, reason given by server: %s\n",
			hostname, dirname, nfs_strerror(status.fhs_status));
		return 0;
	}
#ifdef HAVE_NFSV3
	memcpy(fh->u.nfs2_fh.data, status.fhstatus_u.fhs_fhandle, NFS_FHSIZE);
#else
	memcpy(fh, status.fhstatus_u.fhs_fhandle, NFS_FHSIZE);
#endif
	return 1;
}

/*
 * Get pathconf info from the server
 */
static int
get_pathconf(CLIENT *clnt, const char *hostname, const char *dirname)
{
	struct timeval	timeout;
	enum clnt_stat	clnt_stat;
	ppathcnf	pathconf;

	timeout.tv_sec  = 20;
	timeout.tv_usec = 0;

	/* try to mount hostname:dirname */
	clnt_stat = clnt_call(clnt, MOUNTPROC_PATHCONF,
		(xdrproc_t) xdr_dirpath,  &dirname,
		(xdrproc_t) xdr_ppathcnf, &pathconf,
		timeout);

	if (clnt_stat != RPC_SUCCESS) {
		clnt_perror(clnt, "rpc pathconf failed");
		return 0;
	}

	/* Some versions of Linux rpc.mountd return a name_max value of
	 * _POSIX_NAME_MAX, which is 14 (which makes UNIX better than DOS
	 * by a margin of three bytes). */
	if (pathconf.pc_name_max < 64) {
		fprintf(stderr, "Server is using a buggy Linux mountd - "
				"ignoring PATHCONF info.\n");
		return 1;
	}

	data.namlen = pathconf.pc_name_max;
#ifdef NFS_MOUNT_DEBUG
	printf("got name length %d\n", data.namlen);
#endif
	return 1;
}

/*
 * Obtain NFSv3 file handle from server
 */
static int
get_rootfh_nfs3(CLIENT *clnt, const char *hostname,
			      const char *dirname, struct nfs_fh *fh)
{
#ifdef HAVE_NFSV3
	fhandle3	*fhandle;
	struct timeval	timeout;
	enum clnt_stat	clnt_stat;
	struct mountres3 result;

	timeout.tv_sec  = 20;
	timeout.tv_usec = 0;

	/* try to mount hostname:dirname */
	memset(&result, 0, sizeof(result));
	clnt_stat = clnt_call(clnt, MOUNTPROC3_MNT,
		(xdrproc_t) xdr_dirpath,   &dirname,
		(xdrproc_t) xdr_mountres3, &result,
		timeout);

	if (clnt_stat != RPC_SUCCESS) {
		clnt_perror(clnt, "rpc mount");
		return 0;
	}
	if (result.fhs_status != MNT_OK) {
		/* Fortunately, MNTv3 error codes match NFSv2 errors */
		fprintf(stderr,
			"mount: %s:%s failed, reason given by server: %s\n",
			hostname, dirname, nfs_strerror(result.fhs_status));
		return 0;
	}

	fhandle = &result.mountres3_u.mountinfo.fhandle;
	fh->u.nfs3_fh.size = fhandle->fhandle3_len;
	memcpy(fh->u.nfs3_fh.data, fhandle->fhandle3_val,
				   fhandle->fhandle3_len);
	return 1;
#else
	return 0;
#endif
}

/*
 * This is the NFS mount routine proper
 */
int
nfsmount(const char *spec, const char *node, int *flags,
	     char **extra_opts, char **mount_opts)
{
	struct sockaddr_in nfs_addr;
	struct sockaddr_in mnt_addr;
	char	hostdir[1024], new_opts[1024];
	CLIENT	*mclient = NULL;
	char	*hostname, *dirname, *old_opts, *mounthost = NULL, *s;

	if (strlen(spec) >= sizeof(hostdir)-1) {
		fprintf(stderr, "mount: "
			"excessively long host:dir argument\n");
		goto fail;
	}
	strcpy(hostdir, spec);
	if ((s = (strchr(hostdir, ':')))) {
		hostname = hostdir;
		dirname = s + 1;
		*s = '\0';
	} else {
		fprintf(stderr, "mount: "
			"directory to mount not in host:dir format\n");
		goto fail;
	}

	/* Get the NFS server address */
	if (!get_servaddr(hostname, &nfs_addr))
		goto fail;

	/* add IP address to mtab options for use when unmounting */
	s = inet_ntoa(nfs_addr.sin_addr);
	old_opts = *extra_opts;
	if (!old_opts)
		old_opts = "";
	if (strlen(old_opts) + strlen(s) + 10 >= sizeof(new_opts)) {
		fprintf(stderr, "mount: "
			"excessively long option argument\n");
		goto fail;
	}
	sprintf(new_opts, "%s%saddr=%s",
		old_opts, *old_opts ? "," : "", s);
	*extra_opts = xstrdup(new_opts);

	/* parse options */
	if (!parse_opts(old_opts))
		goto fail;

	*mount_opts = (char *) &data;
	if (*flags & MS_REMOUNT)
		return 0;

	/* If no specific mount server given, use the nfs server */
	if (mounthost) {
		if (!get_servaddr(mounthost, &mnt_addr))
			goto fail;
	} else {
		memcpy(&mnt_addr, &nfs_addr, sizeof(mnt_addr));
	}

	/* create mount deamon client */
	if (nfsvers == 3)
		mntvers = 3;
	/* For NFSv2, first try MNTv2, then MNTv1 */
	if (mntvers == 0) {
		mclient = get_mount_client(&mnt_addr, mntport, mntprog, 2);
		mntvers = mclient? 2 : 1;
	}
	if (mclient == NULL) {
		mclient = get_mount_client(&mnt_addr, mntport, mntprog, mntvers);
		if (mclient == NULL) {
			clnt_pcreateerror("mount clnt_create");
			goto fail;
		}
	}
	mclient->cl_auth = authunix_create_default();

	/* When using MNTv2, try getting the pathconf info */
	if (mntvers == 2) {
		if (!get_pathconf(mclient, hostname, dirname))
			goto fail;
	}

	if (nfsvers == 2) {
		if (!get_rootfh_nfs2(mclient, hostname, dirname, &data.root))
			goto fail;
	} else {
		if (!get_rootfh_nfs3(mclient, hostname, dirname, &data.root))
			goto fail;
	}

	/*
	 * Get the NFS server port.
	 */
	if (nfsport == 0) {
		nfs_addr.sin_port = PMAPPORT;
		nfsport = pmap_getport(&nfs_addr, nfsprog, nfsvers, nfsproto);
		if (nfsport == 0)
			nfsport = NFS_PORT;
#ifdef NFS_MOUNT_DEBUG
		else
			printf("used portmapper to find NFS port\n");
#endif
	}
#ifdef NFS_MOUNT_DEBUG
	printf("using port %d for nfs deamon\n", nfsport);
#endif
	nfs_addr.sin_port = htons(nfsport);

	/*
	 * Prepare data structure for kernel.
	 */
	memcpy(&data.addr, &nfs_addr, sizeof(data.addr));
	strncpy(data.hostname, hostname, sizeof(data.hostname)-1);

	/* clean up */
	auth_destroy(mclient->cl_auth);
	clnt_destroy(mclient);
	return 0;

	/* abort */
fail:
	if (mclient != NULL)
		clnt_destroy(mclient);
	return 1;
}


/*
 * We need to translate between nfs status return values and
 * the local errno values which may not be the same.
 *
 * MNTv1/MNTv2 implementations often return NFSv2 error when the mount
 * call failed (even though the spec only uses true/false). Newer NFS
 * implementations which also support version 3 frequently return a mix
 * of NFSv2/NFSv3 error codes.
 *
 * MNTv3 specifies its own set of error codes, which fortunately match
 * NFSv2 error codes, even though only a subset is specified. However,
 * it throws in a bunch of new NFSv3 errors in the 10000+ range which
 * must be checked separately.
 *
 * The bottom line being: throw it together, stir well and hope that Coda
 * will blow NFS away quickly.
 */

#ifndef EDQUOT
#define EDQUOT	ENOSPC
#endif

static struct {
	int			stat;	/* MNT return value */
	int			errno;	/* system error number */
} nfs_errtbl[] = {
	{ NFS_OK,		0		},
	{ NFSERR_PERM,		EPERM		},
	{ NFSERR_NOENT,		ENOENT		},
	{ NFSERR_IO,		EIO		},
	{ NFSERR_NXIO,		ENXIO		},
	{ NFSERR_ACCES,		EACCES		},
	{ NFSERR_EXIST,		EEXIST		},
	{ NFSERR_NODEV,		ENODEV		},
	{ NFSERR_NOTDIR,	ENOTDIR		},
	{ NFSERR_ISDIR,		EISDIR		},
#ifdef NFSERR_INVAL
	{ NFSERR_INVAL,		EINVAL		},	/* that Sun forgot */
#endif
	{ NFSERR_FBIG,		EFBIG		},
	{ NFSERR_NOSPC,		ENOSPC		},
	{ NFSERR_ROFS,		EROFS		},
	{ NFSERR_NAMETOOLONG,	ENAMETOOLONG	},
	{ NFSERR_NOTEMPTY,	ENOTEMPTY	},
	{ NFSERR_DQUOT,		EDQUOT		},
	{ NFSERR_STALE,		ESTALE		},
#ifdef EWFLUSH
	{ NFSERR_WFLUSH,	EWFLUSH		},
#endif
	/* Throw in some NFSv3 values for even more fun (HP returns these) */
	{ 71,			EREMOTE		},

	{ -1,			EIO		}
};

static char *nfs_strerror(int stat)
{
	int i;
	static char buf[256];

	for (i = 0; nfs_errtbl[i].stat != -1; i++) {
		if (nfs_errtbl[i].stat == stat)
			return strerror(nfs_errtbl[i].errno);
	}

	/* Now check some special MNTv3 error codes that don't have
	 * a errno.h equivalent. */
	switch (stat) {
	case MNT3ERR_NOTSUPP:
		return "operation not supported";
	case MNT3ERR_SERVERFAULT:
		return "server fault";
	}

	sprintf(buf, "unknown mount status return value: %d", stat);
	return buf;
}

